/**
  * ******************************************************
  *                                                       *
  *   Copyright (C) Microsoft. All rights reserved.       *
  *                                                       *
  *******************************************************
  */
"use strict";
const child_process = require("child_process");
const fs = require("fs");
const os = require("os");
const path = require("path");
const readline = require("readline");
const rimraf = require("rimraf");
const utils = require("vcremote-utils");
var resources = utils.ResourcesManager;
var utilHelper = utils.UtilHelper;
var exec = child_process.exec;
var Certs;
(function (Certs) {
    var debug = false;
    let CERT_DEFAULTS;
    (function (CERT_DEFAULTS) {
        CERT_DEFAULTS.days = 365; // 1 year
        CERT_DEFAULTS.country = "US";
    })(CERT_DEFAULTS || (CERT_DEFAULTS = {}));
    ;
    Promise.defer = function () {
        var result = {};
        result.promise = new Promise(function (resolve, reject) {
            result.resolve = resolve;
            result.reject = reject;
        });
        return result;
    };
    function resetServerCert(conf, yesOrNoHandler) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        if (!fs.existsSync(certsDir)) {
            return initializeServerCerts(conf);
        }
        var shouldProceedDeferred = new Promise.defer;
        yesOrNoHandler = yesOrNoHandler || readline.createInterface({ input: process.stdin, output: process.stdout });
        var answerCallback = function (answer) {
            answer = answer.toLowerCase();
            if (answer === "y" || answer === "yes") {
                yesOrNoHandler.close();
                shouldProceedDeferred.resolve(true);
            }
            else if (answer === "n" || answer === "no") {
                yesOrNoHandler.close();
                shouldProceedDeferred.resolve(false);
            }
            else {
                yesOrNoHandler.question(resources.getStringForLanguage(conf.get("lang"), "YesOrNoQuestion") + os.EOL, answerCallback);
            }
        };
        yesOrNoHandler.question(resources.getStringForLanguage(conf.get("lang"), "ResetServerCertQuestion") + os.EOL, answerCallback);
        return shouldProceedDeferred.promise.
            then(function (shouldProceed) {
            if (shouldProceed) {
                rimraf.sync(certsDir);
                return initializeServerCerts(conf);
            }
            return new Promise(function (resolve, reject) { resolve({}); });
        });
    }
    Certs.resetServerCert = resetServerCert;
    ;
    function initializeServerCerts(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        Certs.authorizedClientCertsPath = path.join(certsDir, "Authorized-Clients");
        var certPaths = {
            certsDir: certsDir,
            serverKeyPath: path.join(certsDir, "server-key.pem"),
            serverCertPath: path.join(certsDir, "server-cert.pem"),
            newCerts: false
        };
        var certsExist = fs.existsSync(certPaths.serverKeyPath) && fs.existsSync(certPaths.serverCertPath);
        certPaths.newCerts = !certsExist;
        var promise;
        if (certsExist) {
            promise = isExpired(certPaths.serverCertPath).
                then(function (hasExpired) {
                if (hasExpired) {
                    // Cert has expired so needs making.
                    return new Promise(function (resolve, reject) { resolve(true); });
                }
                else {
                    // Cert still valid, so does not need making.
                    return new Promise(function (resolve, reject) { resolve(false); });
                }
            });
        }
        else {
            promise = new Promise(function (resolve, reject) { resolve(true); }); // do not exist, so true -> need making
        }
        promise = promise.then(function (shouldMake) {
            if (!shouldMake) {
                return new Promise(function (resolve, reject) { resolve({}); });
            }
            utilHelper.createDirectoryIfNecessary(certsDir);
            utilHelper.createDirectoryIfNecessary(Certs.authorizedClientCertsPath);
            var options = certOptionsFromConf(conf);
            return makeSelfSignedCert(certPaths.serverKeyPath, certPaths.serverCertPath, options, conf).
                then(function () {
                printVisualStudioSetupToConsole(conf, certPaths.serverCertPath);
                certPaths.newCerts = true;
            });
        }).then(function () {
            return {
                newCerts: certPaths.newCerts,
                getKey: function () { return fs.readFileSync(certPaths.serverKeyPath); },
                getCert: function () { return fs.readFileSync(certPaths.serverCertPath); },
            };
        });
        return promise;
    }
    Certs.initializeServerCerts = initializeServerCerts;
    function getServerCerts(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        var certPaths = {
            serverKeyPath: path.join(certsDir, "server-key.pem"),
            serverCertPath: path.join(certsDir, "server-cert.pem")
        };
        var certsExist = fs.existsSync(certPaths.serverKeyPath) && fs.existsSync(certPaths.serverCertPath);
        if (!certsExist) {
            return {
                newCerts: false,
                getKey: function () { throw new Error("Not Supported"); },
                getCert: function () { throw new Error("Not Supported"); },
            };
        }
        return {
            newCerts: false,
            getKey: function () { return fs.readFileSync(certPaths.serverKeyPath); },
            getCert: function () { return fs.readFileSync(certPaths.serverCertPath); },
        };
    }
    Certs.getServerCerts = getServerCerts;
    function isExpired(certPath) {
        return displayCert(certPath, ["dates"]).
            then(function (output) {
            var notAfter = new Date(output.stdout.substring(output.stdout.indexOf("notAfter=") + 9, output.stdout.length - 1));
            return (notAfter.getTime() < new Date().getTime());
        });
    }
    Certs.isExpired = isExpired;
    ;
    // display fields an array of any of these: 'subject', 'issuer', 'dates', etc. (see https://www.openssl.org/docs/apps/x509.html)
    function displayCert(certPath, displayFields) {
        // openssl x509 -noout -in selfsigned-cert.pem -subject -issuer -dates
        var args = "x509 -noout -in " + certPath;
        (displayFields || []).forEach(function (f) {
            args += " -" + f;
        });
        return openSslPromise(args);
    }
    Certs.displayCert = displayCert;
    ;
    // Returns fingerprint for given certPath.
    function getFingerprint(certPath) {
        return displayCert(certPath, ["fingerprint"]).
            then(function (output) {
            var shaString = output.stdout.substring("SHA1 Fingerprint=".length, output.stdout.length - 1);
            return shaString;
        });
    }
    Certs.getFingerprint = getFingerprint;
    ;
    function removeAllCertsSync(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        if (fs.existsSync(certsDir)) {
            rimraf.sync(certsDir);
        }
    }
    Certs.removeAllCertsSync = removeAllCertsSync;
    // Makes a new private key and certificate.
    function makeSelfSignedCert(outKeyPath, outCertPath, options, conf) {
        options = options || {};
        var csrPath = path.join(path.dirname(outCertPath), "CSR-" + path.basename(outCertPath));
        var days = options.days || CERT_DEFAULTS.days;
        var cn = "Mac-vcremote-server";
        var cnfPath = path.join(conf.get("serverDir"), "certs", "openssl.cnf");
        writeConfigFile(cnfPath);
        return openSslPromise("genrsa -out " + outKeyPath + " 1024").
            then(function () {
            return openSslPromise("req -new -subj /CN=" + cn + " -key " + outKeyPath + " -out " + csrPath + " -config " + cnfPath);
        }).
            then(function () {
            return openSslPromise("x509 -req -days " + days + " -in " + csrPath + " -signkey " + outKeyPath +
                " -extensions v3_req -extfile " + cnfPath + " -set_serial 01 -out " + outCertPath);
        }).
            then(function () {
            fs.unlinkSync(csrPath);
        });
    }
    Certs.makeSelfSignedCert = makeSelfSignedCert;
    ;
    function verifyCert(caCertPath, certPath) {
        return openSslPromise("verify -CAfile " + caCertPath + " " + certPath);
    }
    Certs.verifyCert = verifyCert;
    ;
    function openSslPromise(args) {
        var deferred = new Promise.defer;
        exec("openssl " + args, function (error, stdout, stderr) {
            if (debug) {
                console.info("exec openssl " + args);
                console.info("stdout: %s", stdout);
                console.info("stderr: %s", stderr);
            }
            if (error) {
                deferred.reject(error);
            }
            else {
                deferred.resolve({ stdout: stdout.toString(), stderr: stderr.toString() });
            }
        });
        return deferred.promise;
    }
    function certOptionsFromConf(conf) {
        var options = {};
        if (conf.get("certExpirationdays") < 1) {
            console.info(resources.getString("CertExpirationInvalid", conf.get("certExpirationdays"), CERT_DEFAULTS.days));
        }
        else {
            options.days = conf.get("certExpirationdays");
        }
        return options;
    }
    function writeConfigFile(cnfPath) {
        var net = os.networkInterfaces();
        var cnf = "[req]\ndistinguished_name = req_distinguished_name\nreq_extensions = v3_req\n[req_distinguished_name]\nC_default = US\n[ v3_req ]\nbasicConstraints = CA:FALSE\nkeyUsage = nonRepudiation, digitalSignature, keyEncipherment\nsubjectAltName = @alt_names\n[alt_names]\nDNS.1 = " + os.hostname() + "\n";
        var count = 1;
        for (var key in net) {
            for (var i = 0; i < net[key].length; i++) {
                if (net[key][i].address && !net[key][i].internal) {
                    cnf += "IP." + count + " = " + net[key][i].address + "\n";
                    count++;
                }
            }
        }
        fs.writeFileSync(cnfPath, cnf);
    }
    function makePfx(caCertPath, keyPath, certPath, outPfxPath, options) {
        options = options || {};
        var name = options.name || "pfxFile";
        return openSslPromise("pkcs12 -export -in " + certPath + " -inkey " + keyPath + " -certfile " + caCertPath + " -out " + outPfxPath +
            " -name \'" + name + "\' -password pass:");
    }
    // TODO: Make this explain how to configure vs-mda-client via CLI?
    function printVisualStudioSetupToConsole(conf, serverCertPath) {
        var host = conf.get("hostname") || os.hostname();
        var port = conf.get("port");
        console.info(resources.getString("OSXCertSetupInformation", host, port, serverCertPath));
        console.info("");
    }
    ;
})(Certs || (Certs = {}));
module.exports = Certs;
